from collections import namedtuple
import torch.nn as nn
from .utils import _isinstance

Config = namedtuple('config', [
    'conv_and_linear_weight', 'conv_and_linear_bias', 'bn_weight', 'bn_bias',
    'bn_mean', 'bn_var', 'bn_out', 'acti'
],
                    defaults=[8, 8, 8, 8, 8, 8, 8, 8])

bn_list = [nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d]
acti_list = [nn.ReLU, nn.ReLU6]
conv_linear_list = [nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.Linear]


def config_bn(proc, config):
    if _isinstance(proc, bn_list):
        proc.weight_bit_width = config.bn_weight
        proc.bias_bit_width = config.bn_bias
        proc.mean_bit_width = config.bn_mean
        proc.var_bit_width = config.bn_var
        proc.bn_bit_width = config.bn_out


def config_acti(proc, config):
    if _isinstance(proc, acti_list):
        proc.acti_bit_width = config.acti


def config_conv_linear(proc, config):
    if _isinstance(proc, conv_linear_list):
        proc.weight_bit_width = config.conv_and_linear_weight
        proc.bias_bit_width = config.conv_and_linear_bias


def config_network(net, name, config, show=False):
    proc_list = list(net._modules.keys())
    if proc_list == []:
        if show:
            print(name, 'is configured')
        config_acti(net, config)
        config_conv_linear(net, config)
        config_bn(net, config)
    else:
        for n in proc_list:
            config_network(net._modules[n], name + '.' + n, config, show=show)



"""
Provide quantilized form of torch.nn.modules.batchnorm 
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from .number import directquant


class BatchNorm2d(nn.BatchNorm2d):
    def __init__(self,
                 num_features,
                 eps=1e-5,
                 momentum=0.1,
                 affine=True,
                 track_running_stats=True,
                 weight_bit_width=8,
                 bias_bit_width=8,
                 mean_bit_width=8,
                 var_bit_width=8,
                 bn_bit_width=8):
        super().__init__(num_features, eps, momentum, affine,
                         track_running_stats)

        self.weight_bit_width = weight_bit_width
        self.bias_bit_width = bias_bit_width
        self.mean_bit_width = mean_bit_width
        self.var_bit_width = var_bit_width
        self.bn_bit_width = bn_bit_width

    def bn_forward(self, input):
        mean = directquant(self.running_mean, self.mean_bit_width)
        var = directquant(self.running_var, self.var_bit_width)
        if self.affine is True:
            weight = directquant(self.weight, self.weight_bit_width)
            bias = directquant(self.bias, self.bias_bit_width)
            output = directquant(
                F.batch_norm(input,
                             running_mean=mean,
                             running_var=var,
                             weight=weight,
                             bias=bias), self.bn_bit_width)
        else:
            output = directquant(
                F.batch_norm(input, running_mean=mean, running_var=var),
                self.bn_bit_width)
        return output

    def forward(self, input):
        return self.bn_forward(input)



"""
Provide quantilized form of torch.nn.modules.activation
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from .number import directquant


class ReLU(nn.ReLU):
    def __init__(self,
                 inplace=False,
                 acti_bit_width=8,
                 retrain=True,
                 quant=False):
        super().__init__(inplace)
        self.acti_bit_width = acti_bit_width

    def relu_forward(self, input):
        return directquant(F.relu(input), self.acti_bit_width)

    def forward(self, input):
        return self.relu_forward(input)


class ReLU6(nn.ReLU6):
    def __init__(self, inplace=False, acti_bit_width=8):
        super().__init__(inplace)
        self.acti_bit_width = acti_bit_width

    def relu6_forward(self, input):
        return directquant(F.relu6(input), self.acti_bit_width)

    def forward(self, input):
        return self.relu6_forward(input)



"""
Basic quantization function types are defined for forward propagation and back propagation.
Shown in the paper's Sec. III.C.
"""

import torch
from torch.autograd import Function
import math
from ._utils import number_to_tensor


class StochasticRound(Function):
    @staticmethod
    def forward(ctx, x):
        x_shape = x.shape
        eps = 1e-8
        x = x.reshape(-1, 1)
        choice = torch.cat((x.floor(), x.ceil()), dim=1)
        weight = torch.cat((x.ceil() - x + eps, x + eps - x.floor()), dim=1)
        sample = torch.multinomial(weight, 1)
        idx = torch.cat((sample == 0, sample == 1), dim=1)
        x = choice[idx].reshape(x_shape)
        return x

    @staticmethod
    def backward(ctx, grad_out):
        return grad_out


sround = StochasticRound.apply


class DirectQuant(Function):
    @staticmethod
    def forward(ctx, x, k):
        return torch.clamp(
            torch.round(x * 2**(k - 1)) / 2**(k - 1), -2**(k - 1) + 1,
            2**(k - 1) - 1)

    @staticmethod
    def backward(ctx, grad_out):
        return grad_out, None


directquant = DirectQuant.apply


class ConstQuant(Function):
    @staticmethod
    def forward(ctx, x, k_dr, k):
        r = 2**torch.round(x.abs().max().log2())
        norm = x / r
        dr = 2**(k_dr - 1)
        sd = torch.clamp(sround(dr * norm), -dr + 1, dr - 1)
        return sd / 2**(k - 1)

    @staticmethod
    def backward(ctx, grad_out):
        return grad_out, None, None


constquant = ConstQuant.apply


class ShiftQuant(Function):
    @staticmethod
    def forward(ctx, x, k):
        r = 2**torch.round(x.abs().max().log2())
        norm = x / r
        dk = 1 / 2**(k - 1)
        sq = r * torch.clamp(directquant(norm, k), -1 + dk, 1 - dk)
        return sq

    @staticmethod
    def backward(ctx, grad_out):
        return grad_out, None


shiftquant = ShiftQuant.apply


class GradOfWeightQuant(Function):
    @staticmethod
    def forward(ctx, x, kgw, kdr):
        ctx.save_for_backward(torch.tensor(kgw), torch.tensor(kdr))
        return x

    @staticmethod
    def backward(ctx, grad_out):
        kgw, kdr = ctx.saved_tensors
        return constquant(grad_out, kdr, kgw), None, None


gwquant = GradOfWeightQuant.apply


class GradOfBNQuant(Function):
    @staticmethod
    def forward(ctx, x, k):
        ctx.save_for_backward(torch.tensor(k))
        return x

    @staticmethod
    def backward(ctx, grad_out):
        k = ctx.saved_tensors[0]
        return directquant(grad_out, k), None


gbnquant = GradOfBNQuant.apply


class ErrorQuant(Function):
    @staticmethod
    def forward(ctx, x, k):
        r = 2**torch.round(x.abs().max().log2())
        sc = r / 2**(k - 1)
        cmp0 = ((x / sc).abs >= 1) * sc * torch.clamp(
            (x / sc).round(), 1 - 2**(k - 1), 2**(k - 1) - 1)
        cmp1 = ((x / sc).abs < 1) * directquant(x / sc, k)
        return cmp1 + cmp0

    @staticmethod
    def backward(ctx, grad_out):
        return grad_out


errquant = ErrorQuant.apply


"""
Provide quantilized form of torch.nn.modules.linear
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from .number import directquant


class Linear(nn.Linear):
    def __init__(self,
                 in_features,
                 out_features,
                 bias=True,
                 weight_bit_width=8,
                 bias_bit_width=16):
        super().__init__(in_features, out_features, bias=bias)
        self.weight_bit_width = weight_bit_width
        self.bias_bit_width = bias_bit_width

    def linear_forward(self, input):
        if self.bias is None:
            bias = None
        else:
            bias = directquant(self.bias, self.bias_bit_width)
        weight = directquant(self.weight, self.weight_bit_width)
        return F.linear(input, weight, bias)

    def forward(self, input):
        return self.linear_forward(input)


if __name__ == '__main__':
    lin = Linear(3, 5)
    x = torch.rand(3, 3)
    print(lin(x).shape)



"""
Provide quantilized form of torch.nn.modules.conv
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from .number import directquant


class Conv2d(nn.Conv2d):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        padding_mode='zeros',
        weight_bit_width=8,
        bias_bit_width=16,
    ):
        super().__init__(in_channels=in_channels,
                         out_channels=out_channels,
                         kernel_size=kernel_size,
                         stride=stride,
                         padding=padding,
                         dilation=dilation,
                         groups=groups,
                         bias=bias,
                         padding_mode=padding_mode)
        self.weight_bit_width = weight_bit_width
        self.bias_bit_width = bias_bit_width

    def conv_forward(self, input):
        if self.bias is None:
            bias = None
        else:
            bias = directquant(self.bias, self.bias_bit_width)
        weight = directquant(self.weight, self.weight_bit_width)
        return F.conv2d(input, weight, bias, self.stride, self.padding,
                        self.dilation, self.groups)

    def forward(self, input):
        return self.conv_forward(input)


if __name__ == '__main__':
    conv = Conv2d(3, 6, 3)
    x = torch.rand(4, 3, 5, 5)
    print(conv(x).shape)
